@@ -1,9 +1,9 @@
 use crate::ctype::{isalpha, isdigit, ispunct, isspace};
-use crate::nodes::TableAlignment;
 use crate::nodes::{
     AstNode, ListDelimType, ListType, NodeCodeBlock, NodeHeading, NodeHtmlBlock, NodeLink,
-    NodeTable, NodeValue,
+    NodeMath, NodeTable, NodeValue, NodeWikiLink,
 };
+use crate::nodes::{NodeList, TableAlignment};
 #[cfg(feature = "shortcodes")]
 use crate::parser::shortcodes::NodeShortCode;
 use crate::parser::Options;
@@ -20,6 +20,14 @@ pub fn format_document<'a>(
     options: &Options,
     output: &mut dyn Write,
 ) -> io::Result<()> {
+    // Formatting an ill-formed AST might lead to invalid output. However, we don't want to pay for
+    // validation in normal workflow. As a middleground, we validate the AST in debug builds. See
+    // https://github.com/kivikakk/comrak/issues/371.
+    #[cfg(debug_assertions)]
+    root.validate().unwrap_or_else(|e| {
+        panic!("The document to format is ill-formed: {:?}", e);
+    });
+
     format_document_with_plugins(root, options, output, &Plugins::default())
 }
 
@@ -39,9 +47,9 @@ pub fn format_document_with_plugins<'a>(
     Ok(())
 }
 
-struct CommonMarkFormatter<'a, 'o> {
+struct CommonMarkFormatter<'a, 'o, 'c> {
     node: &'a AstNode<'a>,
-    options: &'o Options,
+    options: &'o Options<'c>,
     v: Vec<u8>,
     prefix: Vec<u8>,
     column: usize,
@@ -53,6 +61,7 @@ struct CommonMarkFormatter<'a, 'o> {
     in_tight_list_item: bool,
     custom_escape: Option<fn(&'a AstNode<'a>, u8) -> bool>,
     footnote_ix: u32,
+    ol_stack: Vec<usize>,
 }
 
 #[derive(PartialEq, Clone, Copy)]
@@ -63,7 +72,7 @@ enum Escaping {
     Title,
 }
 
-impl<'a, 'o> Write for CommonMarkFormatter<'a, 'o> {
+impl<'a, 'o, 'c> Write for CommonMarkFormatter<'a, 'o, 'c> {
     fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
         self.output(buf, false, Escaping::Literal);
         Ok(buf.len())
@@ -74,8 +83,8 @@ impl<'a, 'o> Write for CommonMarkFormatter<'a, 'o> {
     }
 }
 
-impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
-    fn new(node: &'a AstNode<'a>, options: &'o Options) -> Self {
+impl<'a, 'o, 'c> CommonMarkFormatter<'a, 'o, 'c> {
+    fn new(node: &'a AstNode<'a>, options: &'o Options<'c>) -> Self {
         CommonMarkFormatter {
             node,
             options,
@@ -90,6 +99,7 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             in_tight_list_item: false,
             custom_escape: None,
             footnote_ix: 0,
+            ol_stack: vec![],
         }
     }
 
@@ -326,6 +336,9 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
                 )
                 && self.get_in_tight_list_item(node);
         }
+        let next_is_block = node
+            .next_sibling()
+            .map_or(true, |next| next.data.borrow().value.block());
 
         match node.data.borrow().value {
             NodeValue::Document => (),
@@ -345,7 +358,7 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             NodeValue::Text(ref literal) => {
                 self.format_text(literal.as_bytes(), allow_wrap, entering)
             }
-            NodeValue::LineBreak => self.format_line_break(entering),
+            NodeValue::LineBreak => self.format_line_break(entering, next_is_block),
             NodeValue::SoftBreak => self.format_soft_break(allow_wrap, entering),
             NodeValue::Code(ref code) => {
                 self.format_code(code.literal.as_bytes(), allow_wrap, entering)
@@ -377,6 +390,15 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             NodeValue::FootnoteReference(ref nfr) => {
                 self.format_footnote_reference(nfr.name.as_bytes(), entering)
             }
+            NodeValue::MultilineBlockQuote(..) => self.format_block_quote(entering),
+            NodeValue::Escaped => {
+                // noop - automatic escaping is already being done
+            }
+            NodeValue::Math(ref math) => self.format_math(math, allow_wrap, entering),
+            NodeValue::WikiLink(ref nl) => return self.format_wikilink(nl, entering),
+            NodeValue::Underline => self.format_underline(),
+            NodeValue::SpoileredText => self.format_spoiler(),
+            NodeValue::EscapedTag(ref net) => self.format_escaped_tag(net),
         };
         true
     }
@@ -400,18 +422,35 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
     }
 
     fn format_list(&mut self, node: &'a AstNode<'a>, entering: bool) {
-        if !entering
-            && match node.next_sibling() {
+        let ol_start = match node.data.borrow().value {
+            NodeValue::List(NodeList {
+                list_type: ListType::Ordered,
+                start,
+                ..
+            }) => Some(start),
+            _ => None,
+        };
+
+        if entering {
+            if let Some(start) = ol_start {
+                self.ol_stack.push(start);
+            }
+        } else {
+            if ol_start.is_some() {
+                self.ol_stack.pop();
+            }
+
+            if match node.next_sibling() {
                 Some(next_sibling) => matches!(
                     next_sibling.data.borrow().value,
                     NodeValue::CodeBlock(..) | NodeValue::List(..)
                 ),
                 _ => false,
+            } {
+                self.cr();
+                write!(self, "<!-- end list -->").unwrap();
+                self.blankline();
             }
-        {
-            self.cr();
-            write!(self, "<!-- end list -->").unwrap();
-            self.blankline();
         }
     }
 
@@ -426,11 +465,11 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
         let marker_width = if parent.list_type == ListType::Bullet {
             2
         } else {
-            let list_number = match node.data.borrow().value {
-                NodeValue::Item(ref ni) => ni.start,
-                NodeValue::TaskItem(_) => parent.start,
-                _ => unreachable!(),
-            };
+            let last_stack = self.ol_stack.last_mut().unwrap();
+            let list_number = *last_stack;
+            if entering {
+                *last_stack += 1;
+            }
             let list_delim = parent.delimiter;
             write!(
                 listmarker,
@@ -505,12 +544,13 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             let info = ncb.info.as_bytes();
             let literal = ncb.literal.as_bytes();
 
-            if info.is_empty()
-                && (literal.len() > 2
-                    && !isspace(literal[0])
-                    && !(isspace(literal[literal.len() - 1])
-                        && isspace(literal[literal.len() - 2])))
-                && !first_in_list_item
+            #[allow(clippy::len_zero)]
+            if !(info.len() > 0
+                || literal.len() <= 2
+                || isspace(literal[0])
+                || first_in_list_item
+                || self.options.render.prefer_fenced
+                || isspace(literal[literal.len() - 1]) && isspace(literal[literal.len() - 2]))
             {
                 write!(self, "    ").unwrap();
                 write!(self.prefix, "    ").unwrap();
@@ -566,9 +606,13 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
         }
     }
 
-    fn format_line_break(&mut self, entering: bool) {
+    fn format_line_break(&mut self, entering: bool, next_is_block: bool) {
         if entering {
-            if !self.options.render.hardbreaks {
+            if !self.options.render.hardbreaks && !next_is_block {
+                // If the next element is a block, a backslash means a
+                // literal backslash instead of a line break. In this case
+                // we can just skip the line break since it's meaningless
+                // before a block.
                 write!(self, "\\").unwrap();
             }
             self.cr();
@@ -582,6 +626,8 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
                 && !self.options.render.hardbreaks
             {
                 self.cr();
+            } else if self.options.render.hardbreaks {
+                self.output(&[b'\n'], allow_wrap, Escaping::Literal);
             } else {
                 self.output(&[b' '], allow_wrap, Escaping::Literal);
             }
@@ -655,6 +701,18 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
         write!(self, "^").unwrap();
     }
 
+    fn format_underline(&mut self) {
+        write!(self, "__").unwrap();
+    }
+
+    fn format_spoiler(&mut self) {
+        write!(self, "||").unwrap();
+    }
+
+    fn format_escaped_tag(&mut self, net: &String) {
+        self.output(net.as_bytes(), false, Escaping::Literal);
+    }
+
     fn format_link(&mut self, node: &'a AstNode<'a>, nl: &NodeLink, entering: bool) -> bool {
         if is_autolink(node, nl) {
             if entering {
@@ -677,6 +735,24 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
         true
     }
 
+    fn format_wikilink(&mut self, nl: &NodeWikiLink, entering: bool) -> bool {
+        if entering {
+            write!(self, "[[").unwrap();
+            if self.options.extension.wikilinks_title_after_pipe {
+                self.output(nl.url.as_bytes(), false, Escaping::Url);
+                write!(self, "|").unwrap();
+            }
+        } else {
+            if self.options.extension.wikilinks_title_before_pipe {
+                write!(self, "|").unwrap();
+                self.output(nl.url.as_bytes(), false, Escaping::Url);
+            }
+            write!(self, "]]").unwrap();
+        }
+
+        true
+    }
+
     fn format_image(&mut self, nl: &NodeLink, allow_wrap: bool, entering: bool) {
         if entering {
             write!(self, "![").unwrap();
@@ -696,7 +772,7 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
     fn format_shortcode(&mut self, ne: &NodeShortCode, entering: bool) {
         if entering {
             write!(self, ":").unwrap();
-            self.output(ne.shortcode().as_bytes(), false, Escaping::Literal);
+            self.output(ne.code.as_bytes(), false, Escaping::Literal);
             write!(self, ":").unwrap();
         }
     }
@@ -773,6 +849,31 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             self.write_all(b"]").unwrap();
         }
     }
+
+    fn format_math(&mut self, math: &NodeMath, allow_wrap: bool, entering: bool) {
+        if entering {
+            let literal = math.literal.as_bytes();
+            let start_fence = if math.dollar_math {
+                if math.display_math {
+                    "$$"
+                } else {
+                    "$"
+                }
+            } else {
+                "$`"
+            };
+
+            let end_fence = if start_fence == "$`" {
+                "`$"
+            } else {
+                start_fence
+            };
+
+            self.output(start_fence.as_bytes(), false, Escaping::Literal);
+            self.output(literal, allow_wrap, Escaping::Literal);
+            self.output(end_fence.as_bytes(), false, Escaping::Literal);
+        }
+    }
 }
 
 fn longest_char_sequence(literal: &[u8], ch: u8) -> usize {
