@@ -1,7 +1,8 @@
 //! The HTML renderer for the CommonMark AST, as well as helper functions.
 use crate::ctype::isspace;
 use crate::nodes::{
-    AstNode, ListType, NodeCode, NodeFootnoteDefinition, NodeTable, NodeValue, TableAlignment,
+    AstNode, ListType, NodeCode, NodeFootnoteDefinition, NodeMath, NodeTable, NodeValue,
+    TableAlignment,
 };
 use crate::parser::{Options, Plugins};
 use crate::scanners;
@@ -62,25 +63,24 @@ impl<'w> Write for WriteWithLast<'w> {
     }
 }
 
-/// Converts header Strings to canonical, unique, but still human-readable, anchors.
+/// Converts header strings to canonical, unique, but still human-readable,
+/// anchors.
 ///
-/// To guarantee uniqueness, an anchorizer keeps track of the anchors
-/// it has returned.  So, for example, to parse several MarkDown
-/// files, use a new anchorizer per file.
+/// To guarantee uniqueness, an anchorizer keeps track of the anchors it has
+/// returned; use one per output file.
 ///
 /// ## Example
 ///
 /// ```
-/// use comrak::Anchorizer;
-///
+/// # use comrak::Anchorizer;
 /// let mut anchorizer = Anchorizer::new();
-///
 /// // First "stuff" is unsuffixed.
 /// assert_eq!("stuff".to_string(), anchorizer.anchorize("Stuff".to_string()));
 /// // Second "stuff" has "-1" appended to make it unique.
 /// assert_eq!("stuff-1".to_string(), anchorizer.anchorize("Stuff".to_string()));
 /// ```
 #[derive(Debug, Default)]
+#[doc(hidden)]
 pub struct Anchorizer(HashSet<String>);
 
 impl Anchorizer {
@@ -95,12 +95,9 @@ impl Anchorizer {
     /// resultant anchor unique.
     ///
     /// ```
-    /// use comrak::Anchorizer;
-    ///
+    /// # use comrak::Anchorizer;
     /// let mut anchorizer = Anchorizer::new();
-    ///
     /// let source = "Ticks aren't in";
-    ///
     /// assert_eq!("ticks-arent-in".to_string(), anchorizer.anchorize(source.to_string()));
     /// ```
     pub fn anchorize(&mut self, header: String) -> String {
@@ -129,9 +126,9 @@ impl Anchorizer {
     }
 }
 
-struct HtmlFormatter<'o> {
+struct HtmlFormatter<'o, 'c> {
     output: &'o mut WriteWithLast<'o>,
-    options: &'o Options,
+    options: &'o Options<'c>,
     anchorizer: Anchorizer,
     footnote_ix: u32,
     written_footnote_ix: u32,
@@ -288,7 +285,7 @@ pub fn escape(output: &mut dyn Write, buffer: &[u8]) -> io::Result<()> {
 /// The inclusion of characters like "%" in those which are not escaped is
 /// explained somewhat here:
 ///
-/// https://github.com/github/cmark-gfm/blob/c32ef78bae851cb83b7ad52d0fbff880acdcd44a/src/houdini_href_e.c#L7-L31
+/// <https://github.com/github/cmark-gfm/blob/c32ef78bae851cb83b7ad52d0fbff880acdcd44a/src/houdini_href_e.c#L7-L31>
 ///
 /// In other words, if a CommonMark user enters:
 ///
@@ -364,8 +361,12 @@ where
     Ok(())
 }
 
-impl<'o> HtmlFormatter<'o> {
-    fn new(options: &'o Options, output: &'o mut WriteWithLast<'o>, plugins: &'o Plugins) -> Self {
+impl<'o, 'c: 'o> HtmlFormatter<'o, 'c> {
+    fn new(
+        options: &'o Options<'c>,
+        output: &'o mut WriteWithLast<'o>,
+        plugins: &'o Plugins,
+    ) -> Self {
         HtmlFormatter {
             options,
             output,
@@ -417,6 +418,9 @@ impl<'o> HtmlFormatter<'o> {
                             NodeValue::LineBreak | NodeValue::SoftBreak => {
                                 self.output.write_all(b" ")?;
                             }
+                            NodeValue::Math(NodeMath { ref literal, .. }) => {
+                                self.escape(literal.as_bytes())?;
+                            }
                             _ => (),
                         }
                         plain
@@ -445,6 +449,9 @@ impl<'o> HtmlFormatter<'o> {
                 output.extend_from_slice(literal.as_bytes())
             }
             NodeValue::LineBreak | NodeValue::SoftBreak => output.push(b' '),
+            NodeValue::Math(NodeMath { ref literal, .. }) => {
+                output.extend_from_slice(literal.as_bytes())
+            }
             _ => {
                 for n in node.children() {
                     Self::collect_text(n, output);
@@ -582,71 +589,77 @@ impl<'o> HtmlFormatter<'o> {
             },
             NodeValue::CodeBlock(ref ncb) => {
                 if entering {
-                    self.cr()?;
+                    if ncb.info.eq("math") {
+                        self.render_math_code_block(node, &ncb.literal)?;
+                    } else {
+                        self.cr()?;
 
-                    let mut first_tag = 0;
-                    let mut pre_attributes: HashMap<String, String> = HashMap::new();
-                    let mut code_attributes: HashMap<String, String> = HashMap::new();
-                    let code_attr: String;
+                        let mut first_tag = 0;
+                        let mut pre_attributes: HashMap<String, String> = HashMap::new();
+                        let mut code_attributes: HashMap<String, String> = HashMap::new();
+                        let code_attr: String;
 
-                    let literal = &ncb.literal.as_bytes();
-                    let info = &ncb.info.as_bytes();
+                        let literal = &ncb.literal.as_bytes();
+                        let info = &ncb.info.as_bytes();
 
-                    if !info.is_empty() {
-                        while first_tag < info.len() && !isspace(info[first_tag]) {
-                            first_tag += 1;
-                        }
-
-                        let lang_str = str::from_utf8(&info[..first_tag]).unwrap();
-                        let info_str = str::from_utf8(&info[first_tag..]).unwrap().trim();
-
-                        if self.options.render.github_pre_lang {
-                            pre_attributes.insert(String::from("lang"), lang_str.to_string());
-
-                            if self.options.render.full_info_string && !info_str.is_empty() {
-                                pre_attributes
-                                    .insert(String::from("data-meta"), info_str.trim().to_string());
+                        if !info.is_empty() {
+                            while first_tag < info.len() && !isspace(info[first_tag]) {
+                                first_tag += 1;
                             }
-                        } else {
-                            code_attr = format!("language-{}", lang_str);
-                            code_attributes.insert(String::from("class"), code_attr);
 
-                            if self.options.render.full_info_string && !info_str.is_empty() {
-                                code_attributes
-                                    .insert(String::from("data-meta"), info_str.to_string());
+                            let lang_str = str::from_utf8(&info[..first_tag]).unwrap();
+                            let info_str = str::from_utf8(&info[first_tag..]).unwrap().trim();
+
+                            if self.options.render.github_pre_lang {
+                                pre_attributes.insert(String::from("lang"), lang_str.to_string());
+
+                                if self.options.render.full_info_string && !info_str.is_empty() {
+                                    pre_attributes.insert(
+                                        String::from("data-meta"),
+                                        info_str.trim().to_string(),
+                                    );
+                                }
+                            } else {
+                                code_attr = format!("language-{}", lang_str);
+                                code_attributes.insert(String::from("class"), code_attr);
+
+                                if self.options.render.full_info_string && !info_str.is_empty() {
+                                    code_attributes
+                                        .insert(String::from("data-meta"), info_str.to_string());
+                                }
                             }
                         }
-                    }
 
-                    if self.options.render.sourcepos {
-                        let ast = node.data.borrow();
-                        pre_attributes
-                            .insert("data-sourcepos".to_string(), ast.sourcepos.to_string());
-                    }
-
-                    match self.plugins.render.codefence_syntax_highlighter {
-                        None => {
-                            write_opening_tag(self.output, "pre", pre_attributes)?;
-                            write_opening_tag(self.output, "code", code_attributes)?;
-
-                            self.escape(literal)?;
-
-                            self.output.write_all(b"</code></pre>\n")?
+                        if self.options.render.sourcepos {
+                            let ast = node.data.borrow();
+                            pre_attributes
+                                .insert("data-sourcepos".to_string(), ast.sourcepos.to_string());
                         }
-                        Some(highlighter) => {
-                            highlighter.write_pre_tag(self.output, pre_attributes)?;
-                            highlighter.write_code_tag(self.output, code_attributes)?;
 
-                            highlighter.write_highlighted(
-                                self.output,
-                                match str::from_utf8(&info[..first_tag]) {
-                                    Ok(lang) => Some(lang),
-                                    Err(_) => None,
-                                },
-                                &ncb.literal,
-                            )?;
+                        match self.plugins.render.codefence_syntax_highlighter {
+                            None => {
+                                write_opening_tag(self.output, "pre", pre_attributes)?;
+                                write_opening_tag(self.output, "code", code_attributes)?;
 
-                            self.output.write_all(b"</code></pre>\n")?
+                                self.escape(literal)?;
+
+                                self.output.write_all(b"</code></pre>\n")?
+                            }
+                            Some(highlighter) => {
+                                highlighter.write_pre_tag(self.output, pre_attributes)?;
+                                highlighter.write_code_tag(self.output, code_attributes)?;
+
+                                highlighter.write_highlighted(
+                                    self.output,
+                                    match str::from_utf8(&info[..first_tag]) {
+                                        Ok(lang) => Some(lang),
+                                        Err(_) => None,
+                                    },
+                                    &ncb.literal,
+                                )?;
+
+                                self.output.write_all(b"</code></pre>\n")?
+                            }
                         }
                     }
                 }
@@ -712,22 +725,29 @@ impl<'o> HtmlFormatter<'o> {
                 }
             }
             NodeValue::Text(ref literal) => {
+                // Nowhere to put sourcepos.
                 if entering {
                     self.escape(literal.as_bytes())?;
                 }
             }
             NodeValue::LineBreak => {
+                // Unreliable sourcepos.
                 if entering {
                     self.output.write_all(b"<br")?;
-                    self.render_sourcepos(node)?;
+                    if self.options.render.experimental_inline_sourcepos {
+                        self.render_sourcepos(node)?;
+                    }
                     self.output.write_all(b" />\n")?;
                 }
             }
             NodeValue::SoftBreak => {
+                // Unreliable sourcepos.
                 if entering {
                     if self.options.render.hardbreaks {
                         self.output.write_all(b"<br")?;
-                        self.render_sourcepos(node)?;
+                        if self.options.render.experimental_inline_sourcepos {
+                            self.render_sourcepos(node)?;
+                        }
                         self.output.write_all(b" />\n")?;
                     } else {
                         self.output.write_all(b"\n")?;
@@ -735,9 +755,12 @@ impl<'o> HtmlFormatter<'o> {
                 }
             }
             NodeValue::Code(NodeCode { ref literal, .. }) => {
+                // Unreliable sourcepos.
                 if entering {
                     self.output.write_all(b"<code")?;
-                    self.render_sourcepos(node)?;
+                    if self.options.render.experimental_inline_sourcepos {
+                        self.render_sourcepos(node)?;
+                    }
                     self.output.write_all(b">")?;
                     self.escape(literal.as_bytes())?;
                     self.output.write_all(b"</code>")?;
@@ -760,13 +783,17 @@ impl<'o> HtmlFormatter<'o> {
                 }
             }
             NodeValue::Strong => {
+                // Unreliable sourcepos.
                 let parent_node = node.parent();
-                if parent_node.is_none()
-                    || !matches!(parent_node.unwrap().data.borrow().value, NodeValue::Strong)
+                if !self.options.render.gfm_quirks
+                    || (parent_node.is_none()
+                        || !matches!(parent_node.unwrap().data.borrow().value, NodeValue::Strong))
                 {
                     if entering {
                         self.output.write_all(b"<strong")?;
-                        self.render_sourcepos(node)?;
+                        if self.options.render.experimental_inline_sourcepos {
+                            self.render_sourcepos(node)?;
+                        }
                         self.output.write_all(b">")?;
                     } else {
                         self.output.write_all(b"</strong>")?;
@@ -774,54 +801,82 @@ impl<'o> HtmlFormatter<'o> {
                 }
             }
             NodeValue::Emph => {
+                // Unreliable sourcepos.
                 if entering {
                     self.output.write_all(b"<em")?;
-                    self.render_sourcepos(node)?;
+                    if self.options.render.experimental_inline_sourcepos {
+                        self.render_sourcepos(node)?;
+                    }
                     self.output.write_all(b">")?;
                 } else {
                     self.output.write_all(b"</em>")?;
                 }
             }
             NodeValue::Strikethrough => {
+                // Unreliable sourcepos.
                 if entering {
                     self.output.write_all(b"<del")?;
-                    self.render_sourcepos(node)?;
+                    if self.options.render.experimental_inline_sourcepos {
+                        self.render_sourcepos(node)?;
+                    }
                     self.output.write_all(b">")?;
                 } else {
                     self.output.write_all(b"</del>")?;
                 }
             }
             NodeValue::Superscript => {
+                // Unreliable sourcepos.
                 if entering {
                     self.output.write_all(b"<sup")?;
-                    self.render_sourcepos(node)?;
+                    if self.options.render.experimental_inline_sourcepos {
+                        self.render_sourcepos(node)?;
+                    }
                     self.output.write_all(b">")?;
                 } else {
                     self.output.write_all(b"</sup>")?;
                 }
             }
             NodeValue::Link(ref nl) => {
-                if entering {
-                    self.output.write_all(b"<a")?;
-                    self.render_sourcepos(node)?;
-                    self.output.write_all(b" href=\"")?;
-                    let url = nl.url.as_bytes();
-                    if self.options.render.unsafe_ || !dangerous_url(url) {
-                        self.escape_href(url)?;
+                // Unreliable sourcepos.
+                let parent_node = node.parent();
+
+                if !self.options.parse.relaxed_autolinks
+                    || (parent_node.is_none()
+                        || !matches!(
+                            parent_node.unwrap().data.borrow().value,
+                            NodeValue::Link(..)
+                        ))
+                {
+                    if entering {
+                        self.output.write_all(b"<a")?;
+                        if self.options.render.experimental_inline_sourcepos {
+                            self.render_sourcepos(node)?;
+                        }
+                        self.output.write_all(b" href=\"")?;
+                        let url = nl.url.as_bytes();
+                        if self.options.render.unsafe_ || !dangerous_url(url) {
+                            self.escape_href(url)?;
+                        }
+                        if !nl.title.is_empty() {
+                            self.output.write_all(b"\" title=\"")?;
+                            self.escape(nl.title.as_bytes())?;
+                        }
+                        self.output.write_all(b"\">")?;
+                    } else {
+                        self.output.write_all(b"</a>")?;
                     }
-                    if !nl.title.is_empty() {
-                        self.output.write_all(b"\" title=\"")?;
-                        self.escape(nl.title.as_bytes())?;
-                    }
-                    self.output.write_all(b"\">")?;
-                } else {
-                    self.output.write_all(b"</a>")?;
                 }
             }
             NodeValue::Image(ref nl) => {
+                // Unreliable sourcepos.
                 if entering {
+                    if self.options.render.figure_with_caption {
+                        self.output.write_all(b"<figure>")?;
+                    }
                     self.output.write_all(b"<img")?;
-                    self.render_sourcepos(node)?;
+                    if self.options.render.experimental_inline_sourcepos {
+                        self.render_sourcepos(node)?;
+                    }
                     self.output.write_all(b" src=\"")?;
                     let url = nl.url.as_bytes();
                     if self.options.render.unsafe_ || !dangerous_url(url) {
@@ -835,12 +890,21 @@ impl<'o> HtmlFormatter<'o> {
                         self.escape(nl.title.as_bytes())?;
                     }
                     self.output.write_all(b"\" />")?;
+                    if self.options.render.figure_with_caption {
+                        if !nl.title.is_empty() {
+                            self.output.write_all(b"<figcaption>")?;
+                            self.escape(nl.title.as_bytes())?;
+                            self.output.write_all(b"</figcaption>")?;
+                        }
+                        self.output.write_all(b"</figure>")?;
+                    }
                 }
             }
             #[cfg(feature = "shortcodes")]
             NodeValue::ShortCode(ref nsc) => {
+                // Nowhere to put sourcepos.
                 if entering {
-                    self.output.write_all(nsc.emoji().as_bytes())?;
+                    self.output.write_all(nsc.emoji.as_bytes())?;
                 }
             }
             NodeValue::Table(..) => {
@@ -956,16 +1020,17 @@ impl<'o> HtmlFormatter<'o> {
                 }
             }
             NodeValue::FootnoteReference(ref nfr) => {
+                // Unreliable sourcepos.
                 if entering {
                     let mut ref_id = format!("fnref-{}", nfr.name);
-
-                    self.output.write_all(b"<sup")?;
-                    self.render_sourcepos(node)?;
-
                     if nfr.ref_num > 1 {
                         ref_id = format!("{}-{}", ref_id, nfr.ref_num);
                     }
 
+                    self.output.write_all(b"<sup")?;
+                    if self.options.render.experimental_inline_sourcepos {
+                        self.render_sourcepos(node)?;
+                    }
                     self.output
                         .write_all(b" class=\"footnote-ref\"><a href=\"#fn-")?;
                     self.escape_href(nfr.name.as_bytes())?;
@@ -993,6 +1058,87 @@ impl<'o> HtmlFormatter<'o> {
                     self.output.write_all(b"</li>\n")?;
                 }
             }
+            NodeValue::MultilineBlockQuote(_) => {
+                if entering {
+                    self.cr()?;
+                    self.output.write_all(b"<blockquote")?;
+                    self.render_sourcepos(node)?;
+                    self.output.write_all(b">\n")?;
+                } else {
+                    self.cr()?;
+                    self.output.write_all(b"</blockquote>\n")?;
+                }
+            }
+            NodeValue::Escaped => {
+                // Unreliable sourcepos.
+                if self.options.render.escaped_char_spans {
+                    if entering {
+                        self.output.write_all(b"<span data-escaped-char")?;
+                        if self.options.render.experimental_inline_sourcepos {
+                            self.render_sourcepos(node)?;
+                        }
+                        self.output.write_all(b">")?;
+                    } else {
+                        self.output.write_all(b"</span>")?;
+                    }
+                }
+            }
+            NodeValue::Math(NodeMath {
+                ref literal,
+                display_math,
+                dollar_math,
+                ..
+            }) => {
+                if entering {
+                    self.render_math_inline(node, literal, display_math, dollar_math)?;
+                }
+            }
+            NodeValue::WikiLink(ref nl) => {
+                // Unreliable sourcepos.
+                if entering {
+                    self.output.write_all(b"<a")?;
+                    if self.options.render.experimental_inline_sourcepos {
+                        self.render_sourcepos(node)?;
+                    }
+                    self.output.write_all(b" href=\"")?;
+                    let url = nl.url.as_bytes();
+                    if self.options.render.unsafe_ || !dangerous_url(url) {
+                        self.escape_href(url)?;
+                    }
+                    self.output.write_all(b"\" data-wikilink=\"true")?;
+                    self.output.write_all(b"\">")?;
+                } else {
+                    self.output.write_all(b"</a>")?;
+                }
+            }
+            NodeValue::Underline => {
+                // Unreliable sourcepos.
+                if entering {
+                    self.output.write_all(b"<u")?;
+                    if self.options.render.experimental_inline_sourcepos {
+                        self.render_sourcepos(node)?;
+                    }
+                    self.output.write_all(b">")?;
+                } else {
+                    self.output.write_all(b"</u>")?;
+                }
+            }
+            NodeValue::SpoileredText => {
+                // Unreliable sourcepos.
+                if entering {
+                    self.output.write_all(b"<span")?;
+                    if self.options.render.experimental_inline_sourcepos {
+                        self.render_sourcepos(node)?;
+                    }
+                    self.output.write_all(b" class=\"spoiler\">")?;
+                } else {
+                    self.output.write_all(b"</span>")?;
+                }
+            }
+            NodeValue::EscapedTag(ref net) => {
+                // Nowhere to put sourcepos.
+                self.output.write_all(net.as_bytes())?;
+            }
         }
         Ok(false)
     }
@@ -1034,4 +1180,69 @@ impl<'o> HtmlFormatter<'o> {
         }
         Ok(true)
     }
+
+    // Renders a math dollar inline, `$...$` and `$$...$$` using `<span>` to be similar
+    // to other renderers.
+    fn render_math_inline<'a>(
+        &mut self,
+        node: &'a AstNode<'a>,
+        literal: &String,
+        display_math: bool,
+        dollar_math: bool,
+    ) -> io::Result<()> {
+        let mut tag_attributes: Vec<(String, String)> = Vec::new();
+        let style_attr = if display_math { "display" } else { "inline" };
+        let tag: &str = if dollar_math { "span" } else { "code" };
+
+        tag_attributes.push((String::from("data-math-style"), String::from(style_attr)));
+
+        // Unreliable sourcepos.
+        if self.options.render.experimental_inline_sourcepos && self.options.render.sourcepos {
+            let ast = node.data.borrow();
+            tag_attributes.push(("data-sourcepos".to_string(), ast.sourcepos.to_string()));
+        }
+
+        write_opening_tag(self.output, tag, tag_attributes)?;
+        self.escape(literal.as_bytes())?;
+        write!(self.output, "</{}>", tag)?;
+
+        Ok(())
+    }
+
+    // Renders a math code block, ```` ```math ```` using `<pre><code>`
+    fn render_math_code_block<'a>(
+        &mut self,
+        node: &'a AstNode<'a>,
+        literal: &String,
+    ) -> io::Result<()> {
+        self.cr()?;
+
+        // use vectors to ensure attributes always written in the same order,
+        // for testing stability
+        let mut pre_attributes: Vec<(String, String)> = Vec::new();
+        let mut code_attributes: Vec<(String, String)> = Vec::new();
+        let lang_str = "math";
+
+        if self.options.render.github_pre_lang {
+            pre_attributes.push((String::from("lang"), lang_str.to_string()));
+            pre_attributes.push((String::from("data-math-style"), String::from("display")));
+        } else {
+            let code_attr = format!("language-{}", lang_str);
+            code_attributes.push((String::from("class"), code_attr));
+            code_attributes.push((String::from("data-math-style"), String::from("display")));
+        }
+
+        if self.options.render.sourcepos {
+            let ast = node.data.borrow();
+            pre_attributes.push(("data-sourcepos".to_string(), ast.sourcepos.to_string()));
+        }
+
+        write_opening_tag(self.output, "pre", pre_attributes)?;
+        write_opening_tag(self.output, "code", code_attributes)?;
+
+        self.escape(literal.as_bytes())?;
+        self.output.write_all(b"</code></pre>\n")?;
+
+        Ok(())
+    }
 }
