@@ -4,6 +4,9 @@ mod inlines;
 pub mod shortcodes;
 mod table;
 
+pub mod math;
+pub mod multiline_block_quote;
+
 use crate::adapters::SyntaxHighlighterAdapter;
 use crate::arena_tree::Node;
 use crate::ctype::{isdigit, isspace};
@@ -13,18 +16,20 @@ use crate::nodes::{
     Ast, AstNode, ListDelimType, ListType, NodeCodeBlock, NodeDescriptionItem, NodeHeading,
     NodeHtmlBlock, NodeList, NodeValue,
 };
-use crate::scanners;
+use crate::scanners::{self, SetextChar};
 use crate::strings::{self, split_off_front_matter, Case};
 use derive_builder::Builder;
 use std::cell::RefCell;
 use std::cmp::min;
 use std::collections::HashMap;
-use std::fmt::{Debug, Formatter};
+use std::fmt::{self, Debug, Formatter};
 use std::mem;
 use std::str;
+use std::sync::{Arc, Mutex};
 use typed_arena::Arena;
 
 use crate::adapters::HeadingAdapter;
+use crate::parser::multiline_block_quote::NodeMultilineBlockQuote;
 
 use self::inlines::RefMap;
 
@@ -53,55 +58,6 @@ pub fn parse_document<'a>(
     arena: &'a Arena<AstNode<'a>>,
     buffer: &str,
     options: &Options,
-) -> &'a AstNode<'a> {
-    parse_document_with_broken_link_callback(arena, buffer, options, None)
-}
-
-/// Parse a Markdown document to an AST.
-///
-/// In case the parser encounters any potential links that have a broken reference (e.g `[foo]`
-/// when there is no `[foo]: url` entry at the bottom) the provided callback will be called with
-/// the reference name, and the returned pair will be used as the link destination and title if not
-/// None.
-///
-/// **Note:** The label provided to the callback is the normalized representation of the label as
-/// described in the [GFM spec](https://github.github.com/gfm/#matches).
-///
-/// ```
-/// use comrak::{Arena, parse_document_with_broken_link_callback, format_html, Options};
-/// use comrak::nodes::{AstNode, NodeValue};
-///
-/// # fn main() -> std::io::Result<()> {
-/// // The returned nodes are created in the supplied Arena, and are bound by its lifetime.
-/// let arena = Arena::new();
-///
-/// let root = parse_document_with_broken_link_callback(
-///     &arena,
-///     "# Cool input!\nWow look at this cool [link][foo]. A [broken link] renders as text.",
-///     &Options::default(),
-///     Some(&mut |link_ref: &str| match link_ref {
-///         "foo" => Some((
-///             "https://www.rust-lang.org/".to_string(),
-///             "The Rust Language".to_string(),
-///         )),
-///         _ => None,
-///     }),
-/// );
-///
-/// let mut output = Vec::new();
-/// format_html(root, &Options::default(), &mut output)?;
-/// let output_str = std::str::from_utf8(&output).expect("invalid UTF-8");
-/// assert_eq!(output_str, "<h1>Cool input!</h1>\n<p>Wow look at this cool \
-///                 <a href=\"https://www.rust-lang.org/\" title=\"The Rust Language\">link</a>. \
-///                 A [broken link] renders as text.</p>\n");
-/// # Ok(())
-/// # }
-/// ```
-pub fn parse_document_with_broken_link_callback<'a, 'c>(
-    arena: &'a Arena<AstNode<'a>>,
-    buffer: &str,
-    options: &Options,
-    callback: Option<Callback<'c>>,
 ) -> &'a AstNode<'a> {
     let root: &'a AstNode<'a> = arena.alloc(Node::new(RefCell::new(Ast {
         value: NodeValue::Document,
@@ -111,14 +67,53 @@ pub fn parse_document_with_broken_link_callback<'a, 'c>(
         open: true,
         last_line_blank: false,
         table_visited: false,
+        line_offsets: Vec::with_capacity(0),
     })));
-    let mut parser = Parser::new(arena, root, options, callback);
+    let mut parser = Parser::new(arena, root, options);
     let mut linebuf = Vec::with_capacity(buffer.len());
     parser.feed(&mut linebuf, buffer, true);
     parser.finish(linebuf)
 }
 
-type Callback<'c> = &'c mut dyn FnMut(&str) -> Option<(String, String)>;
+/// Parse a Markdown document to an AST, specifying
+/// [`ParseOptions::broken_link_callback`].
+#[deprecated(
+    since = "0.25.0",
+    note = "The broken link callback has been moved into ParseOptions<'c>."
+)]
+pub fn parse_document_with_broken_link_callback<'a, 'c>(
+    arena: &'a Arena<AstNode<'a>>,
+    buffer: &str,
+    options: &Options<'c>,
+    callback: Option<BrokenLinkCallback<'c>>,
+) -> &'a AstNode<'a> {
+    let mut options_with_callback = options.clone();
+    options_with_callback.parse.broken_link_callback = callback.map(|cb| Arc::new(Mutex::new(cb)));
+    parse_document(arena, buffer, &options_with_callback)
+}
+
+/// The type of the callback used when a reference link is encountered with no
+/// matching reference.
+///
+/// The details of the broken reference are passed in the
+/// [`BrokenLinkReference`] argument. If a [`ResolvedReference`] is returned, it
+/// is used as the link; otherwise, no link is made and the reference text is
+/// preserved in its entirety.
+pub type BrokenLinkCallback<'c> =
+    &'c mut dyn FnMut(BrokenLinkReference) -> Option<ResolvedReference>;
+
+/// Struct to the broken link callback, containing details on the link reference
+/// which failed to find a match.
+#[derive(Debug)]
+pub struct BrokenLinkReference<'l> {
+    /// The normalized reference link label. Unicode case folding is applied;
+    /// see <https://github.com/commonmark/commonmark-spec/issues/695> for a
+    /// discussion on the details of what this exactly means.
+    pub normalized: &'l str,
+
+    /// The original text in the link label.
+    pub original: &'l str,
+}
 
 pub struct Parser<'a, 'o, 'c> {
     arena: &'a Arena<AstNode<'a>>,
@@ -139,19 +134,19 @@ pub struct Parser<'a, 'o, 'c> {
     last_line_length: usize,
     last_buffer_ended_with_cr: bool,
     total_size: usize,
-    options: &'o Options,
-    callback: Option<Callback<'c>>,
+    options: &'o Options<'c>,
 }
 
 #[derive(Default, Debug, Clone)]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-/// Umbrella options struct.
-pub struct Options {
+/// Umbrella options struct. `'c` represents the lifetime of any callback
+/// closure options may take.
+pub struct Options<'c> {
     /// Enable CommonMark extensions.
     pub extension: ExtensionOptions,
 
     /// Configure parse-time options.
-    pub parse: ParseOptions,
+    pub parse: ParseOptions<'c>,
 
     /// Configure render-time options.
     pub render: RenderOptions,
@@ -337,6 +332,73 @@ pub struct ExtensionOptions {
     /// ```
     pub front_matter_delimiter: Option<String>,
 
+    /// Enables the multiline block quote extension.
+    ///
+    /// Place `>>>` before and after text to make it into
+    /// a block quote.
+    ///
+    /// ``` md
+    /// Paragraph one
+    ///
+    /// >>>
+    /// Paragraph two
+    ///
+    /// - one
+    /// - two
+    /// >>>
+    /// ```
+    ///
+    /// ```
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// options.extension.multiline_block_quotes = true;
+    /// assert_eq!(markdown_to_html(">>>\nparagraph\n>>>", &options),
+    ///            "<blockquote>\n<p>paragraph</p>\n</blockquote>\n");
+    /// ```
+    pub multiline_block_quotes: bool,
+
+    /// Enables math using dollar syntax.
+    ///
+    /// ``` md
+    /// Inline math $1 + 2$ and display math $$x + y$$
+    ///
+    /// $$
+    /// x^2
+    /// $$
+    /// ```
+    ///
+    /// ```
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// options.extension.math_dollars = true;
+    /// assert_eq!(markdown_to_html("$1 + 2$ and $$x = y$$", &options),
+    ///            "<p><span data-math-style=\"inline\">1 + 2</span> and <span data-math-style=\"display\">x = y</span></p>\n");
+    /// assert_eq!(markdown_to_html("$$\nx^2\n$$\n", &options),
+    ///            "<p><span data-math-style=\"display\">\nx^2\n</span></p>\n");
+    /// ```
+    pub math_dollars: bool,
+
+    /// Enables math using code syntax.
+    ///
+    /// ```` md
+    /// Inline math $`1 + 2`$
+    ///
+    /// ```math
+    /// x^2
+    /// ```
+    /// ````
+    ///
+    /// ```
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// options.extension.math_code = true;
+    /// assert_eq!(markdown_to_html("$`1 + 2`$", &options),
+    ///            "<p><code data-math-style=\"inline\">1 + 2</code></p>\n");
+    /// assert_eq!(markdown_to_html("```math\nx^2\n```\n", &options),
+    ///            "<pre><code class=\"language-math\" data-math-style=\"display\">x^2\n</code></pre>\n");
+    /// ```
+    pub math_code: bool,
+
     #[cfg(feature = "shortcodes")]
     #[cfg_attr(docsrs, doc(cfg(feature = "shortcodes")))]
     /// Phrases wrapped inside of ':' blocks will be replaced with emojis.
@@ -352,14 +414,105 @@ pub struct ExtensionOptions {
     ///            "<p>Happy Friday! 😄</p>\n");
     /// ```
     pub shortcodes: bool,
+
+    /// Enables wikilinks using title after pipe syntax
+    ///
+    /// ```` md
+    /// [[url|link label]]
+    /// ````
+    ///
+    /// ```
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// options.extension.wikilinks_title_after_pipe = true;
+    /// assert_eq!(markdown_to_html("[[url|link label]]", &options),
+    ///            "<p><a href=\"url\" data-wikilink=\"true\">link label</a></p>\n");
+    /// ```
+    pub wikilinks_title_after_pipe: bool,
+
+    /// Enables wikilinks using title before pipe syntax
+    ///
+    /// ```` md
+    /// [[link label|url]]
+    /// ````
+    ///
+    /// ```
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// options.extension.wikilinks_title_before_pipe = true;
+    /// assert_eq!(markdown_to_html("[[link label|url]]", &options),
+    ///            "<p><a href=\"url\" data-wikilink=\"true\">link label</a></p>\n");
+    /// ```
+    pub wikilinks_title_before_pipe: bool,
+
+    /// Enables underlines using double underscores
+    ///
+    /// ```md
+    /// __underlined text__
+    /// ```
+    ///
+    /// ```
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// options.extension.underline = true;
+    ///
+    /// assert_eq!(markdown_to_html("__underlined text__", &options),
+    ///            "<p><u>underlined text</u></p>\n");
+    /// ```
+    pub underline: bool,
+
+    /// Enables spoilers using double vertical bars
+    ///
+    /// ```md
+    /// Darth Vader is ||Luke's father||
+    /// ```
+    ///
+    /// ```
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// options.extension.spoiler = true;
+    ///
+    /// assert_eq!(markdown_to_html("Darth Vader is ||Luke's father||", &options),
+    ///            "<p>Darth Vader is <span class=\"spoiler\">Luke's father</span></p>\n");
+    /// ```
+    pub spoiler: bool,
+
+    /// Requires at least one space after a `>` character to generate a blockquote,
+    /// and restarts blockquote nesting across unique lines of input
+    ///
+    /// ```md
+    /// >implying implications
+    ///
+    /// > one
+    /// > > two
+    /// > three
+    /// ```
+    ///
+    /// ```
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// options.extension.greentext = true;
+    ///
+    /// assert_eq!(markdown_to_html(">implying implications", &options),
+    ///            "<p>&gt;implying implications</p>\n");
+    ///
+    /// assert_eq!(markdown_to_html("> one\n> > two\n> three", &options),
+    ///            concat!(
+    ///             "<blockquote>\n",
+    ///             "<p>one</p>\n",
+    ///             "<blockquote>\n<p>two</p>\n</blockquote>\n",
+    ///             "<p>three</p>\n",
+    ///             "</blockquote>\n"));
+    /// ```
+    pub greentext: bool,
 }
 
 #[non_exhaustive]
-#[derive(Default, Debug, Clone, Builder)]
+#[derive(Default, Clone, Builder)]
 #[builder(default)]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 /// Options for parser functions.
-pub struct ParseOptions {
+pub struct ParseOptions<'c> {
     /// Punctuation (quotes, full-stops and hyphens) are converted into 'smart' punctuation.
     ///
     /// ```
@@ -391,7 +544,9 @@ pub struct ParseOptions {
     /// Whether or not a simple `x` or `X` is used for tasklist or any other symbol is allowed.
     pub relaxed_tasklist_matching: bool,
 
-    /// Relax parsing of autolinks, allowing links to be detected inside brackets.
+    /// Relax parsing of autolinks, allow links to be detected inside brackets
+    /// and allow all url schemes. It is intended to allow a very specific type of autolink
+    /// detection, such as `[this http://and.com that]` or `{http://foo.com}`, on a best can basis.
     ///
     /// ```
     /// # use comrak::{markdown_to_html, Options};
@@ -405,6 +560,66 @@ pub struct ParseOptions {
     ///            "<p>[<a href=\"https://foo.com\">https://foo.com</a>]</p>\n");
     /// ```
     pub relaxed_autolinks: bool,
+
+    /// In case the parser encounters any potential links that have a broken
+    /// reference (e.g `[foo]` when there is no `[foo]: url` entry at the
+    /// bottom) the provided callback will be called with the reference name,
+    /// both in normalized form and unmodified, and the returned pair will be
+    /// used as the link destination and title if not [`None`].
+    ///
+    /// ```
+    /// # use std::{str, sync::{Arc, Mutex}};
+    /// # use comrak::{Arena, ResolvedReference, parse_document, format_html, Options, BrokenLinkReference, ParseOptionsBuilder};
+    /// # use comrak::nodes::{AstNode, NodeValue};
+    /// #
+    /// # fn main() -> std::io::Result<()> {
+    /// let arena = Arena::new();
+    /// let mut cb = |link_ref: BrokenLinkReference| match link_ref.normalized {
+    ///     "foo" => Some(ResolvedReference {
+    ///         url: "https://www.rust-lang.org/".to_string(),
+    ///         title: "The Rust Language".to_string(),
+    ///     }),
+    ///     _ => None,
+    /// };
+    /// let options = Options {
+    ///     parse: ParseOptionsBuilder::default()
+    ///         .broken_link_callback(Some(Arc::new(Mutex::new(&mut cb))))
+    ///         .build()
+    ///         .unwrap(),
+    ///     ..Default::default()
+    /// };
+    ///
+    /// let root = parse_document(
+    ///     &arena,
+    ///     "# Cool input!\nWow look at this cool [link][foo]. A [broken link] renders as text.",
+    ///     &options,
+    /// );
+    ///
+    /// let mut output = Vec::new();
+    /// format_html(root, &Options::default(), &mut output)?;
+    /// assert_eq!(str::from_utf8(&output).unwrap(),
+    ///            "<h1>Cool input!</h1>\n<p>Wow look at this cool \
+    ///            <a href=\"https://www.rust-lang.org/\" title=\"The Rust Language\">link</a>. \
+    ///            A [broken link] renders as text.</p>\n");
+    /// # Ok(())
+    /// # }
+    #[cfg_attr(feature = "arbitrary", arbitrary(default))]
+    pub broken_link_callback: Option<Arc<Mutex<BrokenLinkCallback<'c>>>>,
+}
+
+impl<'c> fmt::Debug for ParseOptions<'c> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> Result<(), fmt::Error> {
+        let mut struct_fmt = f.debug_struct("ParseOptions");
+        struct_fmt.field("smart", &self.smart);
+        struct_fmt.field("default_info_string", &self.default_info_string);
+        struct_fmt.field("relaxed_tasklist_matching", &self.relaxed_tasklist_matching);
+        struct_fmt.field("relaxed_autolinks", &self.relaxed_autolinks);
+        struct_fmt.field(
+            "broken_link_callback.is_some()",
+            &self.broken_link_callback.is_some(),
+        );
+        struct_fmt.finish()
+    }
 }
 
 #[non_exhaustive]
@@ -521,9 +736,9 @@ pub struct RenderOptions {
 
     /// Set the type of [bullet list marker](https://spec.commonmark.org/0.30/#bullet-list-marker) to use. Options are:
     ///
-    /// * `ListStyleType::Dash` to use `-` (default)
-    /// * `ListStyleType::Plus` to use `+`
-    /// * `ListStyleType::Star` to use `*`
+    /// * [`ListStyleType::Dash`] to use `-` (default)
+    /// * [`ListStyleType::Plus`] to use `+`
+    /// * [`ListStyleType::Star`] to use `*`
     ///
     /// ```rust
     /// # use comrak::{markdown_to_commonmark, Options, ListStyleType};
@@ -542,19 +757,141 @@ pub struct RenderOptions {
     /// ```
     pub list_style: ListStyleType,
 
-    /// Include source position attributes in XML output.
+    /// Include source position attributes in HTML and XML output.
     ///
-    /// Not yet compatible with extension.description_lists.
+    /// Sourcepos information is reliable for all core block items, and most
+    /// extensions. The description lists extension still has issues; see
+    /// <https://github.com/kivikakk/comrak/blob/3bb6d4ce/src/tests/description_lists.rs#L60-L125>.
+    ///
+    /// Sourcepos information is **not** reliable for inlines, and is not
+    /// included in HTML without also setting [`experimental_inline_sourcepos`].
+    /// See <https://github.com/kivikakk/comrak/pull/439> for a discussion.
     ///
     /// ```rust
     /// # use comrak::{markdown_to_commonmark_xml, Options};
     /// let mut options = Options::default();
     /// options.render.sourcepos = true;
-    /// let input = "Hello *world*!";
+    /// let input = "## Hello world!";
     /// let xml = markdown_to_commonmark_xml(input, &options);
-    /// assert!(xml.contains("<emph sourcepos=\"1:7-1:13\">"));
+    /// assert!(xml.contains("<text sourcepos=\"1:4-1:15\" xml:space=\"preserve\">"));
     /// ```
     pub sourcepos: bool,
+
+    /// Include inline sourcepos in HTML output, which is known to have issues.
+    /// See <https://github.com/kivikakk/comrak/pull/439> for a discussion.
+    /// ```rust
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// options.render.sourcepos = true;
+    /// let input = "Hello *world*!";
+    /// assert_eq!(markdown_to_html(input, &options),
+    ///            "<p data-sourcepos=\"1:1-1:14\">Hello <em>world</em>!</p>\n");
+    /// options.render.experimental_inline_sourcepos = true;
+    /// assert_eq!(markdown_to_html(input, &options),
+    ///            "<p data-sourcepos=\"1:1-1:14\">Hello <em data-sourcepos=\"1:7-1:13\">world</em>!</p>\n");
+    /// ```
+    pub experimental_inline_sourcepos: bool,
+
+    /// Wrap escaped characters in a `<span>` to allow any
+    /// post-processing to recognize them.
+    ///
+    /// ```rust
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// let input = "Notify user \\@example";
+    ///
+    /// assert_eq!(markdown_to_html(input, &options),
+    ///            "<p>Notify user @example</p>\n");
+    ///
+    /// options.render.escaped_char_spans = true;
+    /// assert_eq!(markdown_to_html(input, &options),
+    ///            "<p>Notify user <span data-escaped-char>@</span>example</p>\n");
+    /// ```
+    pub escaped_char_spans: bool,
+
+    /// Ignore setext headings in input.
+    ///
+    /// ```rust
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// let input = "setext heading\n---";
+    ///
+    /// assert_eq!(markdown_to_html(input, &options),
+    ///            "<h2>setext heading</h2>\n");
+    ///
+    /// options.render.ignore_setext = true;
+    /// assert_eq!(markdown_to_html(input, &options),
+    ///            "<p>setext heading</p>\n<hr />\n");
+    /// ```
+    pub ignore_setext: bool,
+
+    /// Ignore empty links in input.
+    ///
+    /// ```rust
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// let input = "[]()";
+    ///
+    /// assert_eq!(markdown_to_html(input, &options),
+    ///            "<p><a href=\"\"></a></p>\n");
+    ///
+    /// options.render.ignore_empty_links = true;
+    /// assert_eq!(markdown_to_html(input, &options), "<p>[]()</p>\n");
+    /// ```
+    pub ignore_empty_links: bool,
+
+    /// Enables GFM quirks in HTML output which break CommonMark compatibility.
+    ///
+    /// ```rust
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// let input = "****abcd**** *_foo_*";
+    ///
+    /// assert_eq!(markdown_to_html(input, &options),
+    ///            "<p><strong><strong>abcd</strong></strong> <em><em>foo</em></em></p>\n");
+    ///
+    /// options.render.gfm_quirks = true;
+    /// assert_eq!(markdown_to_html(input, &options),
+    ///            "<p><strong>abcd</strong> <em><em>foo</em></em></p>\n");
+    /// ```
+    pub gfm_quirks: bool,
+
+    /// Prefer fenced code blocks when outputting CommonMark.
+    ///
+    /// ```rust
+    /// # use std::str;
+    /// # use comrak::{Arena, Options, format_commonmark, parse_document};
+    /// let arena = Arena::new();
+    /// let mut options = Options::default();
+    /// let input = "```\nhello\n```\n";
+    /// let root = parse_document(&arena, input, &options);
+    ///
+    /// let mut buf = Vec::new();
+    /// format_commonmark(&root, &options, &mut buf);
+    /// assert_eq!(str::from_utf8(&buf).unwrap(), "    hello\n");
+    ///
+    /// buf.clear();
+    /// options.render.prefer_fenced = true;
+    /// format_commonmark(&root, &options, &mut buf);
+    /// assert_eq!(str::from_utf8(&buf).unwrap(), "```\nhello\n```\n");
+    /// ```
+    pub prefer_fenced: bool,
+
+    /// Render the image as a figure element with the title as its caption.
+    ///
+    /// ```rust
+    /// # use comrak::{markdown_to_html, Options};
+    /// let mut options = Options::default();
+    /// let input = "![image](https://example.com/image.png \"this is an image\")";
+    ///
+    /// assert_eq!(markdown_to_html(input, &options),
+    ///            "<p><img src=\"https://example.com/image.png\" alt=\"image\" title=\"this is an image\" /></p>\n");
+    ///
+    /// options.render.figure_with_caption = true;
+    /// assert_eq!(markdown_to_html(input, &options),
+    ///            "<p><figure><img src=\"https://example.com/image.png\" alt=\"image\" title=\"this is an image\" /><figcaption>this is an image</figcaption></figure></p>\n");
+    /// ```
+    pub figure_with_caption: bool,
 }
 
 #[non_exhaustive]
@@ -623,9 +960,13 @@ impl Debug for RenderPlugins<'_> {
     }
 }
 
-#[derive(Clone)]
-pub struct Reference {
+/// A reference link's resolved details.
+#[derive(Clone, Debug)]
+pub struct ResolvedReference {
+    /// The destination URL of the reference link.
     pub url: String,
+
+    /// The text of the link.
     pub title: String,
 }
 
@@ -636,13 +977,8 @@ struct FootnoteDefinition<'a> {
     total_references: u32,
 }
 
-impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
-    fn new(
-        arena: &'a Arena<AstNode<'a>>,
-        root: &'a AstNode<'a>,
-        options: &'o Options,
-        callback: Option<Callback<'c>>,
-    ) -> Self {
+impl<'a, 'o, 'c: 'o> Parser<'a, 'o, 'c> {
+    fn new(arena: &'a Arena<AstNode<'a>>, root: &'a AstNode<'a>, options: &'o Options<'c>) -> Self {
         Parser {
             arena,
             refmap: RefMap::new(),
@@ -663,7 +999,6 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
             last_buffer_ended_with_cr: false,
             total_size: 0,
             options,
-            callback,
         }
     }
 
@@ -950,7 +1285,8 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
                     }
                 }
                 NodeValue::Table(..) => {
-                    if !table::matches(&line[self.first_nonspace..]) {
+                    if !table::matches(&line[self.first_nonspace..], self.options.extension.spoiler)
+                    {
                         return (false, container, should_continue);
                     }
                     continue;
@@ -963,6 +1299,16 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
                         return (false, container, should_continue);
                     }
                 }
+                NodeValue::MultilineBlockQuote(..) => {
+                    if !self.parse_multiline_block_quote_prefix(
+                        line,
+                        container,
+                        ast,
+                        &mut should_continue,
+                    ) {
+                        return (false, container, should_continue);
+                    }
+                }
                 _ => {}
             }
         }
@@ -970,6 +1316,17 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
         (true, container, should_continue)
     }
 
+    fn is_not_greentext(&mut self, line: &[u8]) -> bool {
+        !self.options.extension.greentext || strings::is_space_or_tab(line[self.first_nonspace + 1])
+    }
+
+    fn setext_heading_line(&mut self, s: &[u8]) -> Option<SetextChar> {
+        match self.options.render.ignore_setext {
+            false => scanners::setext_heading_line(s),
+            true => None,
+        }
+    }
+
     fn open_new_blocks(&mut self, container: &mut &'a AstNode<'a>, line: &[u8], all_matched: bool) {
         let mut matched: usize = 0;
         let mut nl: NodeList = NodeList::default();
@@ -985,7 +1342,27 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
             self.find_first_nonspace(line);
             let indented = self.indent >= CODE_INDENT;
 
-            if !indented && line[self.first_nonspace] == b'>' {
+            if !indented
+                && self.options.extension.multiline_block_quotes
+                && unwrap_into(
+                    scanners::open_multiline_block_quote_fence(&line[self.first_nonspace..]),
+                    &mut matched,
+                )
+            {
+                let first_nonspace = self.first_nonspace;
+                let offset = self.offset;
+                let nmbc = NodeMultilineBlockQuote {
+                    fence_length: matched,
+                    fence_offset: first_nonspace - offset,
+                };
+                *container = self.add_child(
+                    container,
+                    NodeValue::MultilineBlockQuote(nmbc),
+                    self.first_nonspace + 1,
+                );
+                self.advance_offset(line, first_nonspace + matched - offset, false);
+            } else if !indented && line[self.first_nonspace] == b'>' && self.is_not_greentext(line)
+            {
                 let blockquote_startpos = self.first_nonspace;
 
                 let offset = self.first_nonspace + 1 - self.offset;
@@ -1072,7 +1449,7 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
             } else if !indented
                 && node_matches!(container, NodeValue::Paragraph)
                 && unwrap_into(
-                    scanners::setext_heading_line(&line[self.first_nonspace..]),
+                    self.setext_heading_line(&line[self.first_nonspace..]),
                     &mut sc,
                 )
             {
@@ -1255,7 +1632,7 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
 
     fn parse_block_quote_prefix(&mut self, line: &[u8]) -> bool {
         let indent = self.indent;
-        if indent <= 3 && line[self.first_nonspace] == b'>' {
+        if indent <= 3 && line[self.first_nonspace] == b'>' && self.is_not_greentext(line) {
             self.advance_offset(line, indent + 1, true);
 
             if strings::is_space_or_tab(line[self.offset]) {
@@ -1365,7 +1742,7 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
 
     fn parse_html_block_prefix(&mut self, t: u8) -> bool {
         match t {
-            1 | 2 | 3 | 4 | 5 => true,
+            1..=5 => true,
             6 | 7 => !self.blank,
             _ => unreachable!(),
         }
@@ -1444,6 +1821,51 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
         }
     }
 
+    fn parse_multiline_block_quote_prefix(
+        &mut self,
+        line: &[u8],
+        container: &'a AstNode<'a>,
+        ast: &mut Ast,
+        should_continue: &mut bool,
+    ) -> bool {
+        let (fence_length, fence_offset) = match ast.value {
+            NodeValue::MultilineBlockQuote(ref node_value) => {
+                (node_value.fence_length, node_value.fence_offset)
+            }
+            _ => unreachable!(),
+        };
+
+        let matched = if self.indent <= 3 && line[self.first_nonspace] == b'>' {
+            scanners::close_multiline_block_quote_fence(&line[self.first_nonspace..]).unwrap_or(0)
+        } else {
+            0
+        };
+
+        if matched >= fence_length {
+            *should_continue = false;
+            self.advance_offset(line, matched, false);
+
+            // The last child, like an indented codeblock, could be left open.
+            // Make sure it's finalized.
+            if nodes::last_child_is_open(container) {
+                let child = container.last_child().unwrap();
+                let child_ast = &mut *child.data.borrow_mut();
+
+                self.finalize_borrowed(child, child_ast).unwrap();
+            }
+
+            self.current = self.finalize_borrowed(container, ast).unwrap();
+            return false;
+        }
+
+        let mut i = fence_offset;
+        while i > 0 && strings::is_space_or_tab(line[self.offset]) {
+            self.advance_offset(line, 1, true);
+            i -= 1;
+        }
+        true
+    }
+
     fn add_child(
         &mut self,
         mut parent: &'a AstNode<'a>,
@@ -1484,6 +1906,7 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
                     container.first_child().is_some()
                         || container.data.borrow().sourcepos.start.line != self.line_number
                 }
+                NodeValue::MultilineBlockQuote(..) => false,
                 _ => true,
             };
 
@@ -1496,6 +1919,11 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
         if !self.current.same_node(last_matched_container)
             && container.same_node(last_matched_container)
             && !self.blank
+            && (!self.options.extension.greentext
+                || !matches!(
+                    container.data.borrow().value,
+                    NodeValue::BlockQuote | NodeValue::Document
+                ))
             && node_matches!(self.current, NodeValue::Paragraph)
         {
             self.add_line(self.current, line);
@@ -1505,13 +1933,13 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
             }
 
             let add_text_result = match container.data.borrow().value {
-                NodeValue::CodeBlock(..) => AddTextResult::CodeBlock,
+                NodeValue::CodeBlock(..) => AddTextResult::LiteralText,
                 NodeValue::HtmlBlock(ref nhb) => AddTextResult::HtmlBlock(nhb.block_type),
                 _ => AddTextResult::Otherwise,
             };
 
             match add_text_result {
-                AddTextResult::CodeBlock => {
+                AddTextResult::LiteralText => {
                     self.add_line(container, line);
                 }
                 AddTextResult::HtmlBlock(block_type) => {
@@ -1588,6 +2016,11 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
             }
         }
         if self.offset < line.len() {
+            // since whitespace is stripped off the beginning of lines, we need to keep
+            // track of how much was stripped off. This allows us to properly calculate
+            // inline sourcepos during inline processing.
+            ast.line_offsets.push(self.offset);
+
             ast.content
                 .push_str(str::from_utf8(&line[self.offset..]).unwrap());
         }
@@ -1663,7 +2096,7 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
         } else if match ast.value {
             NodeValue::Document => true,
             NodeValue::CodeBlock(ref ncb) => ncb.fenced,
-            NodeValue::Heading(ref nh) => nh.setext,
+            NodeValue::MultilineBlockQuote(..) => true,
             _ => false,
         } {
             ast.sourcepos.end = (self.line_number, self.curline_end_col).into();
@@ -1775,10 +2208,8 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
             self.options,
             content,
             node_data.sourcepos.start.line,
-            node_data.sourcepos.start.column - 1 + node_data.internal_offset,
             &mut self.refmap,
             &delimiter_arena,
-            self.callback.as_mut(),
         );
 
         while subj.parse_inline(node) {}
@@ -1827,7 +2258,7 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
         match node.data.borrow().value {
             NodeValue::FootnoteDefinition(ref nfd) => {
                 map.insert(
-                    strings::normalize_label(&nfd.name, Case::DontPreserve),
+                    strings::normalize_label(&nfd.name, Case::Fold),
                     FootnoteDefinition {
                         ix: None,
                         node,
@@ -1853,7 +2284,7 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
         let mut replace = None;
         match ast.value {
             NodeValue::FootnoteReference(ref mut nfr) => {
-                let normalized = strings::normalize_label(&nfr.name, Case::DontPreserve);
+                let normalized = strings::normalize_label(&nfr.name, Case::Fold);
                 if let Some(ref mut footnote) = map.get_mut(&normalized) {
                     let ix = match footnote.ix {
                         Some(ix) => ix,
@@ -1936,7 +2367,7 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
                                 }
                             }
                         }
-                        NodeValue::Link(..) | NodeValue::Image(..) => {
+                        NodeValue::Link(..) | NodeValue::Image(..) | NodeValue::WikiLink(..) => {
                             this_bracket = true;
                             break;
                         }
@@ -2030,10 +2461,8 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
             self.options,
             content,
             0, // XXX -1 in upstream; never used?
-            0,
             &mut self.refmap,
             &delimiter_arena,
-            self.callback.as_mut(),
         );
 
         let mut lab: String = match subj.link_label() {
@@ -2085,9 +2514,9 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
             }
         }
 
-        lab = strings::normalize_label(&lab, Case::DontPreserve);
+        lab = strings::normalize_label(&lab, Case::Fold);
         if !lab.is_empty() {
-            subj.refmap.map.entry(lab).or_insert(Reference {
+            subj.refmap.map.entry(lab).or_insert(ResolvedReference {
                 url: String::from_utf8(strings::clean_url(url)).unwrap(),
                 title: String::from_utf8(strings::clean_title(&title)).unwrap(),
             });
@@ -2097,7 +2526,7 @@ impl<'a, 'o, 'c> Parser<'a, 'o, 'c> {
 }
 
 enum AddTextResult {
-    CodeBlock,
+    LiteralText,
     HtmlBlock(u8),
     Otherwise,
 }
@@ -2251,20 +2680,15 @@ pub enum AutolinkType {
     Email,
 }
 
-#[derive(Debug, Clone, Copy)]
+#[derive(Debug, Clone, Copy, Default)]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-/// Options for bulleted list redering in markdown. See `link_style` in [RenderOptions] for more details.
+/// Options for bulleted list redering in markdown. See `link_style` in [`RenderOptions`] for more details.
 pub enum ListStyleType {
     /// The `-` character
+    #[default]
     Dash = 45,
     /// The `+` character
     Plus = 43,
     /// The `*` character
     Star = 42,
 }
-
-impl Default for ListStyleType {
-    fn default() -> Self {
-        ListStyleType::Dash
-    }
-}
